@@ -31,10 +31,10 @@ static bool digital_gain_0db_limit = true;
 
 static int snd_rpi_justboom_dac_init(struct snd_soc_pcm_runtime *rtd)
 {
-	struct snd_soc_codec *codec = rtd->codec;
-	snd_soc_update_bits(codec, PCM512x_GPIO_EN, 0x08, 0x08);
-	snd_soc_update_bits(codec, PCM512x_GPIO_OUTPUT_4, 0xf, 0x02);
-	snd_soc_update_bits(codec, PCM512x_GPIO_CONTROL_1, 0x08,0x08);
+	struct snd_soc_component *component = rtd->codec_dai->component;
+	snd_soc_component_update_bits(component, PCM512x_GPIO_EN, 0x08, 0x08);
+	snd_soc_component_update_bits(component, PCM512x_GPIO_OUTPUT_4, 0xf, 0x02);
+	snd_soc_component_update_bits(component, PCM512x_GPIO_CONTROL_1, 0x08,0x08);
 
 	if (digital_gain_0db_limit)
 	{
@@ -51,15 +51,15 @@ static int snd_rpi_justboom_dac_init(struct snd_soc_pcm_runtime *rtd)
 
 static int snd_rpi_justboom_dac_startup(struct snd_pcm_substream *substream) {
 	struct snd_soc_pcm_runtime *rtd = substream->private_data;
-	struct snd_soc_codec *codec = rtd->codec;
-	snd_soc_update_bits(codec, PCM512x_GPIO_CONTROL_1, 0x08,0x08);
+	struct snd_soc_component *component = rtd->codec_dai->component;
+	snd_soc_component_update_bits(component, PCM512x_GPIO_CONTROL_1, 0x08,0x08);
 	return 0;
 }
 
 static void snd_rpi_justboom_dac_shutdown(struct snd_pcm_substream *substream) {
 	struct snd_soc_pcm_runtime *rtd = substream->private_data;
-	struct snd_soc_codec *codec = rtd->codec;
-	snd_soc_update_bits(codec, PCM512x_GPIO_CONTROL_1, 0x08,0x00);
+	struct snd_soc_component *component = rtd->codec_dai->component;
+	snd_soc_component_update_bits(component, PCM512x_GPIO_CONTROL_1, 0x08,0x00);
 }
 
 /* machine stream operations */
@@ -115,7 +115,7 @@ static int snd_rpi_justboom_dac_probe(struct platform_device *pdev)
 			pdev->dev.of_node, "justboom,24db_digital_gain");
 	}
 
-	ret = snd_soc_register_card(&snd_rpi_justboom_dac);
+	ret = devm_snd_soc_register_card(&pdev->dev, &snd_rpi_justboom_dac);
 	if (ret && ret != -EPROBE_DEFER)
 		dev_err(&pdev->dev,
 			"snd_soc_register_card() failed: %d\n", ret);
@@ -123,11 +123,6 @@ static int snd_rpi_justboom_dac_probe(struct platform_device *pdev)
 	return ret;
 }
 
-static int snd_rpi_justboom_dac_remove(struct platform_device *pdev)
-{
-	return snd_soc_unregister_card(&snd_rpi_justboom_dac);
-}
-
 static const struct of_device_id snd_rpi_justboom_dac_of_match[] = {
 	{ .compatible = "justboom,justboom-dac", },
 	{},
@@ -141,7 +136,6 @@ static struct platform_driver snd_rpi_justboom_dac_driver = {
 		.of_match_table = snd_rpi_justboom_dac_of_match,
 	},
 	.probe          = snd_rpi_justboom_dac_probe,
-	.remove         = snd_rpi_justboom_dac_remove,
 };
 
 module_platform_driver(snd_rpi_justboom_dac_driver);
